%:-initialization(horpod).

horpod:-
	initialize,
	repeat,
	read(T),
	treatall(T),!,
	g_read(typeprec,Ltp),
        buildPrec(Ltp,FullTPrec),
	g_assign(typeprec,FullTPrec),
	g_read(funcprec,Lfp),
        buildPrec(Lfp,FullFPrec),
	g_assign(funcprec,FullFPrec),
	checkParams,
	g_read(rules,Lrules),
	reverse(Lrules,Rrules),
	g_assign(rules,Rrules),!,
	g_assign(crules,1),
	g_assign(termin,1),
	checkSN,!.

horpo:-
	initialize,
	write('Input file: '),
	read(Name),
	makefilename(Name,FName),
	see(FName),
	repeat,
	read(T),
	treatall(T),
	seen,!,
	g_read(typeprec,Ltp),
        buildPrec(Ltp,FullTPrec),
	g_assign(typeprec,FullTPrec),
	g_read(funcprec,Lfp),
        buildPrec(Lfp,FullFPrec),
	g_assign(funcprec,FullFPrec),
	checkParams,
	g_read(rules,Lrules),
	reverse(Lrules,Rrules),
	g_assign(rules,Rrules),!,
	g_assign(crules,1),
	g_assign(termin,1),
	checkSN,
        told,!.

initialize:-
        told,
	seen,
	op(600, xfy, ->),
	op(600, xfy, =),
	op(600, xfy, >),
	g_assign(noproof,0),
	g_assign(typecons,[]),
	g_assign(funcdecs,[]),
	g_assign(rules,[]),
	g_assign(typevars,[]),
	g_assign(typeprec,[]),
	g_assign(typelex,[[],[]]),
	g_assign(funcprec,[]),
	g_assign(funclex,[[],[]]).

makefilename(Name,Name):- 
    file_exists(Name).
makefilename(Name,FName):- 
    name(Name,Lname),
    name('.hor',Lext),
    append(Lname,Lext,Lfile),
    name(FName,Lfile),
    file_exists(FName).
makefilename(_,_):- 
    write('Error: File does not exist'),nl,nl,!,fail.

treatall(end_of_file).
treatall(T):-
    treat(T), fail.

treat(outfile(F)):-!,
        tell(F).
treat(noproof):-!,
	g_assign(noproof,1).
treat(typevars(LT)):-!,
        g_read(typevars,LT1),
        append(LT,LT1,LT2),
        sort(LT2,LT3),
	g_assign(typevars,LT3).
%treat(typecons(S,A)):-
%	nonvar(S),
%        atom(S),!,
%        op(600, fx, S),
%	g_read(typecons,L),
%	g_assign(typecons,[[S,A]|L]).
treat(typecons(S,A)):-!,
	nonvar(S),
	g_read(typecons,L),
	g_assign(typecons,[[S,A]|L]).
%treat(funcdec(F,IL,O)):-
%	nonvar(F),
%        atom(F),!,
%        op(600, fx, F),
%	g_read(funcdecs,L),
%	g_assign(funcdecs,[[F,IL,O]|L]).
treat(funcdec(F,IL,O)):-!,
	nonvar(F),
	g_read(funcdecs,L),
	g_assign(funcdecs,[[F,IL,O]|L]).
treat(rule(Env,Left,Right)):-!,
	g_read(rules,L),
	g_assign(rules,[[Env,Left,Right]|L]).
treat(typeprec(TP)):-!,
	g_read(typeprec,TP1),
        append(TP1,TP,TP2),
	g_assign(typeprec,TP2).
treat(typelex(LRLT,RLLT)):-!,
	g_read(typelex,[LRLT1,RLLT1]),
        append(LRLT,LRLT1,LRLT2),
        append(RLLT,RLLT1,RLLT2),
	g_assign(typelex,[LRLT2,RLLT2]).
treat(funcprec(FP)):-!,
	g_read(funcprec,FP1),
        append(FP1,FP,FP2),
	g_assign(funcprec,FP2).
treat(funclex(LRLF,RLLF)):-!,
	g_read(funclex,[LRLF1,RLLF1]),
        append(LRLF,LRLF1,LRLF2),
        append(RLLF,RLLF1,RLLF2),
	g_assign(funclex,[LRLF2,RLLF2]).
treat(T):-
        write('Input format error.'),nl,
        write(T),nl,!,fail.


buildPrec([],[]).
buildPrec([P1|LPrec],FullPrec):-
	buildPrec(LPrec,FullPrec1),
	add2Prec(P1,FullPrec1,FullPrec).

add2Prec(C>D,FullPrec1,FullPrec):-
	atomic(C),atomic(D),!,
	addgr(C,D,FullPrec1,FullPrec).
add2Prec(C=D,FullPrec1,FullPrec):-
	atomic(C),atomic(D),!,
	addeq(C,D,FullPrec1,FullPrec2),
	addeq(D,C,FullPrec2,FullPrec).
add2Prec(C>D,FullPrec1,FullPrec):-
	atomic(C),
    (D=..['=',E|_];D=..['>',E|_]),
	atomic(E),!,
	add2Prec(D,FullPrec1,FullPrec2),
	addgr(C,E,FullPrec2,FullPrec).
add2Prec(C=D,FullPrec1,FullPrec):-
	atomic(C),
	D=..[_,E|_],
	atomic(E),!,
	add2Prec(D,FullPrec1,FullPrec2),
	addeq(C,E,FullPrec2,FullPrec3),
	addeq(E,C,FullPrec3,FullPrec).

addgr(C,D,FullPrec1,FullPrec):-
    (member([D,Leq,Lgr],FullPrec1);[Leq,Lgr]=[[],[]]),
    append(Leq,Lgr,Ltot),
    addgr1(C,[D|Ltot],FullPrec1,FullPrec),!.

addgr1(C,Ltot,[],[[C,[],Ltot]]).
addgr1(C,Ltot,[[C,Leq,Lgr]|FullPrec1],[[C,Leq,Lgr2]|FullPrec]):-!,
    append(Ltot,Lgr,Lgr1), sort(Lgr1,Lgr2),
    addgr2(C,Ltot,FullPrec1,FullPrec).
addgr1(C,Ltot,[[E,Leq,Lgr]|FullPrec1],[[E,Leq,Lgr2]|FullPrec]):-
    (member(C,Leq);member(C,Lgr)),!,
    append(Ltot,Lgr,Lgr1), sort(Lgr1,Lgr2),
    addgr1(C,Ltot,FullPrec1,FullPrec).
addgr1(C,Ltot,[S|FullPrec1],[S|FullPrec]):-
    addgr1(C,Ltot,FullPrec1,FullPrec).

addgr2(_,_,[],[]).
addgr2(C,Ltot,[[E,Leq,Lgr]|FullPrec1],[[E,Leq,Lgr2]|FullPrec]):-
    (member(C,Leq);member(C,Lgr)),!,
    append(Ltot,Lgr,Lgr1), sort(Lgr1,Lgr2),
    addgr2(C,Ltot,FullPrec1,FullPrec).
addgr2(C,Ltot,[S|FullPrec1],[S|FullPrec]):-
    addgr2(C,Ltot,FullPrec1,FullPrec).

addeq(C,D,FullPrec1,FullPrec):-
    (member([D,Leq,Lgr],FullPrec1);[Leq,Lgr]=[[],[]]),
    Leq1=[D|Leq],
    addgr1(C,Lgr,FullPrec1,FullPrec2),
    addeq1(C,Leq1,FullPrec2,FullPrec),!.


addeq1(_,_,[],[]).
addeq1(C,Ltot,[[C,Leq,Lgr]|FullPrec1],[[C,Leq2,Lgr]|FullPrec]):-!,
    append(Ltot,Leq,Leq1), sort(Leq1,Leq2),
    addeq2(C,Ltot,FullPrec1,FullPrec).
addeq1(C,Ltot,[[E,Leq,Lgr]|FullPrec1],[[E,Leq2,Lgr]|FullPrec]):-
    member(C,Leq),!,
    append(Ltot,Leq,Leq1), sort(Leq1,Leq2),
    addeq1(C,Ltot,FullPrec1,FullPrec).
addeq1(C,Ltot,[[E,Leq,Lgr]|FullPrec1],[[E,Leq,Lgr2]|FullPrec]):-
    member(C,Lgr),!,
    append(Ltot,Lgr,Lgr1), sort(Lgr1,Lgr2),
    addgr1(C,Ltot,FullPrec1,FullPrec).
addeq1(C,Ltot,[S|FullPrec1],[S|FullPrec]):-
    addeq1(C,Ltot,FullPrec1,FullPrec).

addeq2(_,_,[],[]).
addeq2(C,Ltot,[[E,Leq,Lgr]|FullPrec1],[[E,Leq2,Lgr]|FullPrec]):-
    member(C,Leq),!,
    append(Ltot,Leq,Leq1), sort(Leq1,Leq2),
    addeq2(C,Ltot,FullPrec1,FullPrec).
addeq2(C,Ltot,[[E,Leq,Lgr]|FullPrec1],[[E,Leq,Lgr2]|FullPrec]):-
    member(C,Lgr),!,
    append(Ltot,Lgr,Lgr1), sort(Lgr1,Lgr2),
    addeq2(C,Ltot,FullPrec1,FullPrec).
addeq2(C,Ltot,[S|FullPrec1],[S|FullPrec]):-
    addeq2(C,Ltot,FullPrec1,FullPrec).


checkParams:-
	g_read(typevars,L),
	member(X,L), var(X),
	write('Type variabels cannot start with capital letters'),nl,nl,!,fail.
checkParams:-
	g_read(typeprec,TP),
	checkPrecreflex(TP),
	write('The type precedence is not well-founded'),nl,nl,!,fail.
checkParams:-
	g_read(funcprec,FP),
	checkPrecreflex(FP),
	write('The function precedence is not well-founded'),nl,nl,!,fail.
checkParams:-
	g_read(typeprec,TP),
	g_read(typelex,[LRLT,RLLT]),
	\+(checkPrecStat(TP,LRLT,RLLT)),
	write('The type precedence is not compatible with the type statuses'),nl,nl,!,fail.
checkParams:-
	g_read(funcprec,FP),
	g_read(funclex,[LRLF,RLLF]),
	\+(checkPrecStat(FP,LRLF,RLLF)),
	write('The function precedence is not compatible with the function statuses'),nl,nl,!,fail.

checkParams.

checkPrecreflex([[C,_,Lgr]|_]):-
    member(C,Lgr),!.
checkPrecreflex([_|L]):-
    checkPrecreflex(L).

checkPrecStat([],_,_).
checkPrecStat([[G,Leq,_]|R],LRLT,RLLT):-
	checkPrecStat1(G,Leq,LRLT,RLLT),!,
	checkPrecStat(R,LRLT,RLLT).

checkPrecStat1(G,Leq,LRLT,_):-
    member(G,LRLT),!,
    subseteq(Leq,LRLT).
checkPrecStat1(G,Leq,_,RLLT):-
    member(G,RLLT),!,
    subseteq(Leq,RLLT).
checkPrecStat1(_,Leq,LRLT,RLLT):-
    append(LRLT,RLLT,LT),
    disjoint(Leq,LT).

subseteq([],_).
subseteq([X|L1],L2):-
    member(X,L2),
    subseteq(L1,L2).

disjoint([],_):-!.
disjoint([X|L1],L2):-
    \+(member(X,L2)),
    disjoint(L1,L2).

checkSN:-
    g_read(rules,[]),
    g_read(termin,0),
    write('Not all rules could be checked.'),nl,!.
checkSN:-
    g_read(rules,[]),
    g_read(termin,1),
    write('All rules checked. The system is SN'),nl,!.
checkSN:-
    g_read(rules,[[Env,Left,Right]|LR]),
    g_read(crules,N),
    write('Checking rule '), write(N), 
    \+(\+((numbervars([Env,Left,Right]),
	   g_assign(ruleenv,Env),
	   g_assign(ruleleft,Left),
	   g_assign(ruleright,Right)))),
    showrule,
    checkRule(Env,Left,Right),
    g_assign(rules,LR),
    N1 is N+1,
    g_assign(crules,N1),!,
    checkSN.

showrule:-
    g_read(noproof,0),    
    write(':'),nl,
    g_read(ruleenv,IEnv),
    g_read(ruleleft,ILeft), 
    g_read(ruleright,IRight),
    write(IEnv),write(' |-- '),
    write_lamterm(ILeft), write(' --> '), 
    write_lamterm(IRight),nl,nl.
showrule:-
    write(':'),nl,nl.


checkRule(Env,Left,Right):-
    write('  Checking left-hand side type... '),
    checkType(Env,Left,LeftT:LT),
    showresultok(LeftT:LT),
    checkRule1(Env,LeftT:LT,Right).
checkRule(_,_,_):-
    showresultfail, 
    g_assign(termin,0),
    write('  The left-hand side is not typable.'),nl,nl.

showresultok(LeftT:LT):- 
    g_read(noproof,0),
    g_read(ruleenv,IEnv),
    g_read(ruleleft,ILeft),
    write(IEnv),write(' |-- '),
    write_lamterm(ILeft), nl,
    addtypes(ILeft,LeftT:LT,ILeftTW),
    write('  Resulting typed term:  '),
    write_lamterm(ILeftTW),nl,nl.
showresultok(_):- nl.

showresultfail:- 
    g_read(noproof,0),
    g_read(ruleenv,IEnv),
    g_read(ruleleft,ILeft),
    write(IEnv),write(' |-- '),
    write_lamterm(ILeft), nl.
showresultfail:- nl.

checkRule1(Env,LeftT:LT,Right):-
    write('  Checking right-hand side type... '),
    checkType(Env,Right,RightT:RT),
    checkRule2(LeftT:LT,RightT:RT).
checkRule1(_,_,_):-
    showresult1fail,
    g_assign(termin,0),
    write('  The rigth-hand side is not typable.'),nl,nl.

showresult1fail:- 
    g_read(noproof,0),
    g_read(ruleenv,IEnv),
    g_read(ruleright,IRight),
    write(IEnv),write(' |-- '),
    write_lamterm(IRight),nl.
showresult1fail:- nl.

checkRule2(LeftT:LT,RightT:RT):-
    subsumes([[RT,LT]],Sol),
    replacetype1term(RightT:RT,Sol,IRightT:IRT),
    showresult2ok(IRightT:IRT), 
    checkRule3(LeftT:LT,IRightT:IRT).
checkRule2(_,RightT:RT):-
    showresult2fail(RightT:RT), 
    g_assign(termin,0),
    write('  The types of the left-hand side and rigth-hand side are not compatible.'),nl,nl.

showresult2ok(IRightT:IRT):- 
    g_read(noproof,0),
    g_read(ruleenv,IEnv),
    g_read(ruleright,IRight),
    write(IEnv),write(' |-- '),
    write_lamterm(IRight),nl,
    addtypes(IRight,IRightT:IRT,IRightW),
    write('  Resulting typed term:  '),
    write_lamterm(IRightW),nl,nl.
showresult2ok(_):- nl,nl.

showresult2fail(RightT:RT):-
    g_read(noproof,0),
    g_read(ruleenv,IEnv),
    g_read(ruleright,IRight),
    write(IEnv),write(' |-- '),
    write_lamterm(IRight),nl,
    addtypes(IRight,RightT:RT,IRightW),
    write('  Resulting typed term:  '),
    write_lamterm(IRightW),nl,nl.
showresult2fail(_):- nl.


checkRule3(LeftT:LT,IRightT:IRT):-
    write('  Checking ordering... '),nl,
    flush_output,!,
    checkruleord(LeftT:LT,IRightT:IRT,P),
    writeproof(P).

writeproof(fail):-
	g_assign(termin,0),
	write('The rule cannot be proved terminating.'),nl,nl.
writeproof(P):-
        g_read(noproof,0),
	write('   The rule is terminating.'),nl,
	write('   Proof:'),nl,
	prettyprint(3,P),nl,nl.
writeproof(_):-
	write('   The rule is terminating.'),nl,nl.

checkType(Env,X,X:T):-
    var(X),!,
    memberenv(X:T,Env).
checkType(Env,X,X:T):-
    atomic(X),
    memberenv(X:T,Env).
%checkType(_,X,X:T):-
%    atomic(X),!,
%    g_read(funcdecs,LFD),
%    member([X,[],T],LFD).
checkType(Env,lam(X:T1,U),lam(X1,V1:T2):(T1->T2)):-
    memberenv(X:_,Env),!,
    (nonvar(X)-> (name(X,L),append(L,[49],L1),name(X1,L1));true),
    replacebvar(X,X1,U,U1),
    checkType([X1:T1|Env],U1,V1:T2).
checkType(Env,lam(X:T1,U),lam(X,V:T2):(T1->T2)):-!,
    checkType([X:T1|Env],U,V:T2).
checkType(Env,lam(X,U),lam(X1,V1:T2):(T1->T2)):-
    memberenv(X:_,Env),!,
    (nonvar(X)-> (name(X,L),append(L,[49],L1),name(X1,L1));true),
    replacebvar(X,X1,U,U1),
    checkType([X1:T1|Env],U1,V1:T2).
checkType(Env,lam(X,U),lam(X,V:T2):(T1->T2)):-!,
    checkType([X:T1|Env],U,V:T2).
checkType(Env,Term,Term1:T):-
    Term=..[@,U|Args],!,
    checkType(Env,U,U1:T1),
    checkTypelist(Env,Args,Args1),
    length(Args,N),
    checkInstArrow(N,T1,Args1,IT1,IArgs1),
    getOutType(N,IT1,T),
    Term1=..[@,U1:T1|IArgs1].
checkType(Env,Term,Term1:T):-!,
    Term=..[F|Args],
    length(Args,NA),
    g_read(funcdecs,LFD),
    member([F,IL,O],LFD),
    length(IL,NA),
    checkTypelist(Env,Args,Args1),
    insttypelist(IL,Args1,O,IArgs1,T),
    Term1=..[F|IArgs1].
checkType(_,_,_):-
	write('The term can not be typed. Fail.'),nl,nl,!,fail.

checkTypelist(_Env,[],[]).
checkTypelist(Env,[A1|Args],[A2|Args1]):-
    checkType(Env,A1,A2),
    checkTypelist(Env,Args,Args1).

memberenv(X:T1,[Y:T2|_]):-
    X==Y,!,
    eqType(T1,T2),!.
memberenv(X:T,[_|Env]):-
    memberenv(X:T,Env).

membervar(X,[Y|_]):-
    X==Y,!.
membervar(X,[_|L]):-
    membervar(X,L).

getOutType(0,T,T):-!.
getOutType(N,T,O):-
    N1 is N -1,
    T=..[->,_,T1],!,
    getOutType(N1,T1,O).

getInTypes(0,_,[]):-!.
getInTypes(N,T,[I|R]):-
    N1 is N -1,
    T=..[->,I,T1],!,
    getInTypes(N1,T1,R).

getTypeslist([],[]).
getTypeslist([_:T1|Args],[T1|IL1]):-
	getTypeslist(Args,IL1).

replacebvar(X,X1,Y,X1):-X==Y,!.
replacebvar(_,_,Y,Y):-var(Y).
replacebvar(X,_,lam(Y,U),lam(Y,U)):-X==Y,!.
replacebvar(X,X1,lam(Y,U),lam(Y,U1)):-!,
    replacebvar(X,X1,U,U1).
replacebvar(X,X1,U,U1):-
    U=..[F|Args],
    replacebvarlist(X,X1,Args,Args1),
    U1=..[F|Args1].  

replacebvarlist(_,_,[],[]).
replacebvarlist(X,X1,[U|Args],[U1|Args1]):-
    replacebvar(X,X1,U,U1),
    replacebvarlist(X,X1,Args,Args1).

checkInstArrow(N,T,Args,IT,IArgs):-
    g_read(typevars,LTV),
    g_read(typecons,LTC),
    getTypeslist(Args,IAL),
    getInTypes(N,T,TL),
    makeeqs(TL,IAL,Prob),
    mgu(Prob,LTV,LTC,Sol),
    replacetype1termlist(Args,Sol,IArgs),
    replacetype1(T,Sol,IT).

insttypelist(IL,Args,O,IArgs,T):-
    g_read(typevars,LTV),
    g_read(typecons,LTC),
    getTypeslist(Args,IAL),
    makeeqs(IL,IAL,Prob),
    mgu(Prob,LTV,LTC,Sol),
    replacetype1termlist(Args,Sol,IArgs),
    replacetype1(O,Sol,T).

makeeqs([],[],[]).
makeeqs([T1|L1],[T2|L2],[T11=T12|Prob]):-
    sort0([T1,T2],[T11,T12]),
    makeeqs(L1,L2,Prob).

mgu([],_,_,[]).
mgu(Prob,LTV,LTC,Sol):-
    delete1(T=T,Prob,Prob1),!,
    mgu(Prob1,LTV,LTC,Sol),!.
mgu(Prob,_,LTC,_):-
    delete1(T1=T2,Prob,_),
    T1=..[C1|_],
    membereqsol([C1,_],[[->,_]|LTC]),
    T2=..[C2|_],
    membereqsol([C2,_],[[->,_]|LTC]),
    \+(C1=C2),!,fail.
mgu(Prob,LTV,LTC,Sol):-
    delete1((T11->T12)=(T21->T22),Prob,Prob1),!,
    mgu([T11=T21,T12=T22|Prob1],LTV,LTC,Sol),!.
mgu(Prob,LTV,LTC,Sol):-
    delete1(T1=T2,Prob,Prob1),
    T1=..[C|LT1],
    membereqsol([C,_],LTC),
    T2=..[C|LT2],!,
    makeeqs(LT1,LT2,Prob2),
    append(Prob2,Prob1,Prob3),
    mgu(Prob3,LTV,LTC,Sol),!.
mgu(Prob,LTV,LTC,[[T2,T1]|Sol]):-
    delete1(T1=T2,Prob,Prob1),
    membereq(T1,LTV),
    membereq(T2,LTV),!,
    replacetype(Prob1,[T2,T1],Prob2),
    mgu(Prob2,LTV,LTC,Sol),!.
mgu(Prob,LTV,LTC,Sol):-
    delete1(T1=T2,Prob,Prob1),
    membereq(T1,LTV),
    \+(membereq(T2,LTV)),
    (delete(T1=T3,Prob1,Prob2);
     delete(T3=T1,Prob1,Prob2)),
    sort0([T2,T3],[A1,A2]),
    mgu([T1=A1,A1=A2|Prob2],LTV,LTC,Sol),!.
mgu(Prob,LTV,LTC,Sol):-
    delete1(T2=T1,Prob,Prob1),
    membereq(T1,LTV),
    \+(membereq(T2,LTV)),
    (delete(T1=T3,Prob1,Prob2);
     delete(T3=T1,Prob1,Prob2)),
    sort0([T2,T3],[A1,A2]),
    mgu([T1=A1,A1=A2|Prob2],LTV,LTC,Sol),!.
mgu(Prob,LTV,_,Sol):-
    mgusol(Prob,LTV,Sol),!.

mgusol([],_,[]).
mgusol([T2=T1|Prob],LTV,[[T1,T2]|Sol1]):-
    membereq(T1,LTV),!,
    mgusol(Prob,LTV,Sol),!,
    replacetype1(T2,Sol,T3),!,
    \+(subt(T1,T3)),
    replaceinsol(T1,T2,Sol,Sol1).
mgusol([T1=T2|Prob],LTV,[[T1,T2]|Sol1]):-
    mgusol(Prob,LTV,Sol),!,
    replacetype1(T2,Sol,T3),!,
    \+(subt(T1,T3)),
    replaceinsol(T1,T2,Sol,Sol1).


subt(X,X):-!.
subt(X,T):-
    T=..[_|Args],!,
    subtlist(X,Args),!.

subtlist(X,[T|_]):-
    subt(X,T),!.
subtlist(X,[_|Args]):-
    subtlist(X,Args),!.

subsumes(Prob,Sol):-
    g_read(typevars,LTV),
    g_read(typecons,LTC),
    match(Prob,LTV,LTC,Sol1),
    sort(Sol1,Sol),
    matchsol(Sol).

match([],_,_,[]).
match([[T,T]|Prob],LTV,LTC,Sol):-
    match(Prob,LTV,LTC,Sol).
match([[(T11->T12),(T21->T22)]|Prob],LTV,LTC,Sol):-
    match([[T11,T21],[T12,T22]|Prob],LTV,LTC,Sol).
match([[T1,T2]|Prob],LTV,LTC,Sol):-
    T1=..[C|LT1],
    membereqsol([C,_],LTC),
    T2=..[C|LT2],!,
    makeeqs(LT1,LT2,Prob1),
    append(Prob1,Prob,Prob2),
    match(Prob2,LTV,LTC,Sol).
match([[T1,T2]|Prob],LTV,LTC,[[T1,T2]|Sol]):-
    membereq(T1,LTV),!,
    match(Prob,LTV,LTC,Sol).

matchsol([]).
matchsol([_]).
matchsol([[X,_],[Y,T2]|L]):-
    X\=Y,!,
    matchsol([[Y,T2]|L]).

replacetype1term(T:I,Sol,T:I1):-
	   var(T),!,
	   replacetype1(I,Sol,I1).
replacetype1term(T:I,Sol,T1:I1):-
	   replacetype1(I,Sol,I1),
	   T=..[O|Args],
	   replacetype1termlist(Args,Sol,Args1),
	   T1=..[O|Args1].

replacetype1termlist([],_,[]).
replacetype1termlist([T:I|RT],Sol,[T1:I1|RT1]):-
    replacetype1term(T:I,Sol,T1:I1),
    replacetype1termlist(RT,Sol,RT1).
replacetype1termlist([X|RT],Sol,[X|RT1]):-
    atom(X),
    replacetype1termlist(RT,Sol,RT1).

replacetype([],_,[]).
replacetype([I|IL],Sol,[I1|IL1]):-
    replacetype1(I,Sol,I1),
    replacetype(IL,Sol,IL1).

replaceinsol(_,_,[],[]).
replaceinsol(T1,T2,[[V,I]|IL],[[V,I1]|IL1]):-
    replacetype1(I,[T1,T2],I1),
    replaceinsol(T1,T2,IL,IL1).

replacetype1(I,Sol,I1):-
    membereqsol([I,I1],Sol),!.
replacetype1(I,_,I):-
    var(I),!.
replacetype1(I,_,I):-
    atomic(I),!.
replacetype1(I,Sol,I1):-
    I=..[C|AI],
    replacetype(AI,Sol,AI1),
    I1=..[C|AI1].

membereq(X,[Y|_]):-
    X==Y,!.
membereq(X,[_|L]):-
    membereq(X,L).

membereq2(X,[[Y|_]|_]):-
    X==Y,!.
membereq2(X,[_|L]):-
    membereq2(X,L).

membereqsol([X,Z],[[Y,Z]|_]):-
    X==Y,!.
membereqsol(P,[_|L]):-
    membereqsol(P,L).

addtypes('$VAR'(N),_:T,'$VAR'(N):T):-!.
addtypes(lam(X,U),lam(_,V):T,lam(X,W):T):-!,
	addtypes(U,V,W).
addtypes(IT,TT:T,ITT:T):-
	IT=..[F|ArgsI],
	TT=..[F|ArgsT],
	addtypeslist(ArgsI,ArgsT,ArgsIT),
	ITT=..[F|ArgsIT].

addtypeslist([],[],[]).
addtypeslist([X|R1],[Y|R2],[Z|R3]):-
	addtypes(X,Y,Z),
	addtypeslist(R1,R2,R3).

checkruleord(Left:T1,Right:T2,Proof):-
	checkhorpoType(Left:T1,Right:T2,[],Proof).
checkruleord(_,_,fail).


checkhorpoType(Left:T1,Right:T2,X,Proof):-
    geType(T1,T2),!,
    checkhorpo(Left:T1,Right:T2,X,Proof).

checkhorpogeType(A1,A2,equal):-
    eqlamTerm(A1,A2).
checkhorpogeType(A1,A2,X,Proof):-
    checkhorpoType(A1,A2,X,Proof).


checkhorpo(Left:T1,Right:T2,X,Proof):-
    nonvar(Left),
    checkhorpo1(Left,T1,Right,T2,X,Proof).

checkhorpoge(Left,_,Right,_,_,equal):-
    eqlamTerm(Left,Right).
checkhorpoge(Left,T1,Right,T2,X,Proof):-
    checkhorpoType(Left,T1,Right,T2,X,Proof).


checkhorpo1(_,_,Right,_,X,p(4-a)):-
    membervar(Right,X),!.
checkhorpo1(Left,T1,Right,_,X,p(4-b,P)):-
    \+(Left=lam(_,_)),
    nonvar(Right),
    Right=lam(Y,U),
    replace(Y,Z,U,V:Tv),
    checkhorpo1(Left,T1,V,Tv,[Z|X],P),!.
checkhorpo1(Left,T1,Right,T2,X,P):-
    Left=..[F|Args],
    g_read(funcdecs,LFD),
    membereq2(F,LFD),!,
    checkhorpo1Func(F,Args,T1,Right,T2,X,P).
checkhorpo1(Left,T1,Right,T2,X,P):-
    Left=..[@|Args],!,
    makeapp2(Args,T1,@(A1,A2)),
    checkhorpo1App([A1,A2],T1,Right,T2,X,P).
checkhorpo1(lam(X,U),T1,Right,T2,VX,P):-!,
    checkhorpo1Lam(lam(X,U),T1,Right,T2,VX,P).

checkhorpo1Func(_,_,_,Right,_,X,p(4-a)):-
    membervar(Right,X),!.
checkhorpo1Func(_F,Args,_,Right,T2,_,p(1-a,N,equal)):- %Case 1a Subterm arg
    nth(N,Args,A),
    eqlamTerm(A,Right:T2),!.
checkhorpo1Func(_F,Args,_,Right,T2,X,p(1-a,N,subtacc)):- %Case 1a Subterm acc
    remXapp(Right:T2,NR:NT2,X),
    nth(N,Args,A),
    buildsubtacceq(A,NR:NT2),!.
%    buildsubtacceq(A,SA:NT2),
%    eqlamTerm(SA:NT2,NR:NT2),!.
checkhorpo1Func(F,Args,T1,Right,_,X,p(1-c,LP)):- %Case 1c FuncApp
    nonvar(Right),
    Right=..[@|Args1],
    checkhorpo1FuncList(F,Args,T1,Args1,X,LP),!.
checkhorpo1Func(F,Args,T1,Right,_,X,p(4-b,P)):- %Case 4b FuncLam
    nonvar(Right),
    Right=lam(Y,U),
    replace(Y,Z,U,V:Tv),
    checkhorpo1Func(F,Args,T1,V,Tv,[Z|X],P),!.
checkhorpo1Func(F,Args,T1,Right,_,X,p(1-c,LP)):- %Case 1c Prec 
    nonvar(Right),
    Right=..[F1|Args1],F1 \== @, F1 \== lam,
    g_read(funcdecs,LFD),
    membereq2(F1,LFD),
    gtPrecF(F,F1),
    checkhorpo1FuncList(F,Args,T1,Args1,X,LP),!.
checkhorpo1Func(F,Args,T1,Right,_,X,p(S,LP)):- %Case 2b Mul %Case 2b Lex
    nonvar(Right),
    Right=..[F1|Args1],F1 \== @, F1 \== lam,
    g_read(funcdecs,LFD),
    membereq2(F1,LFD),
    eqPrecT(F,F1),
    checkhorpo1FuncStat(F,Args,T1,Args1,S,X,LP),!.
checkhorpo1Func(_,Args,_,Right,T2,_X,p(1-a,N,P)):- %Case 1 Subterm acc
    nth(N,Args,A),
    checkhorpoType(A,Right:T2,[],P),!.


checkhorpo1FuncStat(C,Args1,T,Args2,1-b-lr,X,LP):-
    g_read(funclex,[L1,_]),
    member(C,L1),!,
    checkhorpoLex(1,Args1,Args2,X,LP1),
    checkhorpo1FuncList(C,Args1,T,Args2,X,LP2),
    append(LP1,[args,LP2],LP),!.
checkhorpo1FuncStat(C,Args1,T,Args2,1-b-rl,X,LP):-
    g_read(funclex,[_,L2]),
    member(C,L2),!,
    reverse(Args1,IArgs1),
    reverse(Args2,IArgs2),
    checkhorpoLex(0,IArgs1,IArgs2,X,LP1),
    checkhorpo1FuncList(C,Args1,T,Args2,X,LP2),
    append(LP1,[args,LP2],LP),!.
checkhorpo1FuncStat(C,Args1,T,Args2,1-b-mul,X,LP):-
    checkhorpoMulAcc(1,Args1,Args2,X,LP1),
    checkhorpo1FuncList(C,Args1,T,Args2,X,LP2),
    append(LP1,[args,LP2],LP),!.


checkhorpoLex(N,[A1|Args1],[A2|Args2],X,[p(N,equal)|LP]):-
    N1 is N+1,
    eqlamTerm(A1,A2),!,
    checkhorpoLex(N1,Args1,Args2,X,LP),!.
checkhorpoLex(N,[A1|_],[A2|_],X,[p(N,subtacc)]):-
    remXapp(A2,NA2:NT2,X),
    buildsubtacceq(A1,NA2:NT2),!.
%    buildsubtacceq(A1,SA1:NT2),
%    eqlamTerm(SA1:NT2,NA2:NT2),!.
checkhorpoLex(N,[A1|_],[A2|_],_,[p(N,P)]):-
    checkhorpoType(A1,A2,[],P),!.

checkhorpoMul(_,L,[],_,[]):-member(X,L),X\=[],!.
checkhorpoMul(N,Args1,[A2|Args2],X,[p(N1,N,equal)|LP]):-
    nthdelete1(N1,A1,Args1,Args3),
    eqlamTerm(A1,A2),!,
    N2 is N+1,
    checkhorpoMul(N2,Args3,Args2,X,LP).
checkhorpoMul(N,Args1,[A2|Args2],X,[p(N1,N,P)|LP]):-
    nth(N1,Args1,A1),
    checkhorpoType(A1,A2,X,P),!,
    N2 is N+1,
    checkhorpoeqMul(N2,Args1,Args2,X,LP).

checkhorpoeqMul(_,_,[],_,[]).
checkhorpoeqMul(N,Args1,[A2|Args2],X,[p(N1,N,equal)|LP]):-
    nthdelete1(N1,A1,Args1,Args3),
    eqlamTerm(A1,A2),!,
    N2 is N+1,
    checkhorpoeqMul(N2,Args3,Args2,X,LP).
checkhorpoeqMul(N,Args1,[A2|Args2],X,[p(N1,N,P)|LP]):-
    nth(N1,Args1,A1),
    checkhorpoType(A1,A2,X,P),!,
    N2 is N+1,
    checkhorpoeqMul(N2,Args1,Args2,X,LP).

checkhorpoMulAcc(_,L,[],_,[]):-member(X,L),X\=[],!.
checkhorpoMulAcc(N,Args1,[A2|Args2],X,[p(N1,N,equal)|LP]):-
    nthdelete1(N1,A1,Args1,Args3),
    eqlamTerm(A1,A2),!,
    N2 is N+1,
    checkhorpoMulAcc(N2,Args3,Args2,X,LP).
checkhorpoMulAcc(N,Args1,[A2|Args2],X,[p(N1,N,subtacc)|LP]):-
    remXapp(A2,NA2:NT2,X),
    nth(N1,Args1,A1),
    buildsubtacceq(A1,NA2:NT2),!,
%    buildsubtacceq(A1,SA1:NT2),
%    eqlamTerm(SA1:NT2,NA2:NT2),!.
    N2 is N+1,
    checkhorpoeqMulAcc(N2,Args1,Args2,X,LP).
checkhorpoMulAcc(N,Args1,[A2|Args2],X,[p(N1,N,P)|LP]):-
    nth(N1,Args1,A1),
    checkhorpoType(A1,A2,[],P),!,
    N2 is N+1,
    checkhorpoeqMulAcc(N2,Args1,Args2,X,LP).

checkhorpoeqMulAcc(_,_,[],_,[]).
checkhorpoeqMulAcc(N,Args1,[A2|Args2],X,[p(N1,N,equal)|LP]):-
    nthdelete1(N1,A1,Args1,Args3),
    eqlamTerm(A1,A2),!,
    N2 is N+1,
    checkhorpoeqMulAcc(N2,Args3,Args2,X,LP).
checkhorpoeqMulAcc(N,Args1,[A2|Args2],X,[p(N1,N,subtacc)|LP]):-
    remXapp(A2,NA2:NT2,X),
    nthdelete1(N1,A1,Args1,Args3),
    buildsubtacceq(A1,NA2:NT2),!,
%    buildsubtacceq(A1,SA1:NT2),
%    eqlamTerm(SA1:NT2,NA2:NT2),!.
    N2 is N+1,
    checkhorpoeqMulAcc(N2,Args3,Args2,X,LP).
checkhorpoeqMulAcc(N,Args1,[A2|Args2],X,[p(N1,N,P)|LP]):-
    nth(N1,Args1,A1),
    checkhorpoType(A1,A2,[],P),!,
    N2 is N+1,
    checkhorpoeqMulAcc(N2,Args1,Args2,X,LP).

checkhorpo1FuncList(_,_,_,[],_,[]). 
checkhorpo1FuncList(C,Args1,T1,[A2:T2|Args2],X,[p(P)|LP]):-
	checkhorpo1Func(C,Args1,T1,A2,T2,X,P),
	checkhorpo1FuncList(C,Args1,T1,Args2,X,LP).

notin(X,Y):-
	(var(Y);atomic(Y)),!,
	X \== Y,!.
notin(X,lam(Y,_)):-
	X == Y,!.
notin(X,lam(_,U:_)):-!,
	notin(X,U),!.
notin(X,T):-!,
	T=..[_|Args],!,
	notinlist(X,Args),!.

notinlist(_,[]).
notinlist(X,[A:_|Args]):-
	notin(X,A),!,
	notinlist(X,Args).

checkhorpo1App(Args,_,Right,T2,_,p(2-a,N,equal)):- %Case 2a SubtApp arg
    nth(N,Args,A),
    eqlamTerm(A,Right:T2),!.
%checkhorpo1App(Args,_,Right,T2,X,p(2-a,N,subtacc)):- %Case 2a SubtApt acc
%    nth(N,Args,A),
%    remXapp(Right:T2,NR:NT2,X),
%    buildsubtacceq(A,NR:NT2),!.
% Incorrect!!!!
%    buildsubtacceq(A,SA:NT2),
%    eqlamTerm(SA:NT2,NR:NT2),!.
checkhorpo1App(Args,_,Right,T2,_X,p(2-b,pf(2),LP)):- %Case 2b MulApp
    nonvar(Right),
    Right=..[@|Args2],
    makeapp2(Args2,T2,@(A1,A2)),
    checkhorpoMul(1,Args,[A1,A2],[],LP),!.
%    checkhorpoMul(1,Args,[A1,A2],X,LP),!.
checkhorpo1App(Args,_,Right,T2,_,p(2-a,N,P)):- %Case 2a SubtApp
    nth(N,Args,A),
    checkhorpoType(A,Right:T2,[],P),!.
checkhorpo1App([A1,A2],T1,Right,T2,X,p(2-c,P)):- %Case 2c beta
    nonvar(A1),
    A1=lam(X,U),!,
    replace(X,A2,U,W),
    checkhorpoge(W:T1,Right:T2,X,P),!.
checkhorpo1App([A1,A2],_,Right,T2,X,p(2-d,p(2-b,pf(2),LP))):- %Case 2d interm Beta
        nonvar(Right),
        Right=..[F|Args1],
        g_read(funcdecs,LFD),
        membereq2(F,LFD),
        append(Args11,[Ult:Tu],Args1),
        append(Args11,[X:Tu],AArgs1),
        ARight=..[F|AArgs1],
        checkhorpoMul(1,[A1,A2],[lam(X,ARight:T2):(Tu->T2),Ult:Tu],X,LP),!.
        
makeapp2(Args,T1,@(A1:(Tf->T1),Af:Tf)):-
	append(Args1,[Af:Tf],Args),
	makeapp(Args1,A1).
makeapp([A1:_],A1):-!.
makeapp(Args1,A1):-
	A1=..[@|Args1],!.

extracttype([_:T1],T2,(T1->T2)):-!.
extracttype([_:T1|R2],T2,(T1->T)):-
	extracttype(R2,T2,T).

replace(X,T,Y:O,T:O):-
	X==Y,!.
replace(_,_,U:O,U:O):-
	(var(U);atomic(U)),!.
replace(X,_,lam(Y,V):O,lam(Y,V):O):-
	X==Y,!.
replace(X,T,lam(Y,V):O,lam(Y,W):O):-
	replace(X,T,V,W),!.
replace(X,T,U:O,W:O):-
	U=..[F|Args],
	replacelist(X,T,Args,Args1),
	W=..[F|Args1],!.

replacelist(_,_,[],[]).
replacelist(X,T,[A|Args],[A1|Args1]):-
	replace(X,T,A,A1),
	replacelist(X,T,Args,Args1).

checkhorpo1Lam(lam(X,U),_,Right,T2,_,p(3-a,equal)):-  %Case  3a SubtLam
	replace(X,_,U,V),
	eqlamTerm(V,Right:T2),!.
checkhorpo1Lam(lam(X,U),_,Right,T2,XV,p(3-a,P)):-  %Case  3a SubtLam
	replace(X,_,U,V),
	checkhorpoType(V,Right:T2,XV,P).
checkhorpo1Lam(lam(X,U),(TI1->_),Right,T2,XV,p(3-b,P)):-  %Case 3b MonLam
	nonvar(Right),
	nonvar(T2),
	Right=lam(Y,V),
	T2=(TI2->_),
	eqType(TI1,TI2),!,
	replace(Y,X,V,W),
	checkhorpo(U,W,XV,P).
checkhorpo1Lam(lam(X,U),T1,Right,T2,XV,p(3-c,P)):-  %Case 3c Eta
	nonvar(U),
	U=..[@|Args],
	append(Args,Args1,[Last]),
	Last==U,
	notinlist(X,Args1),!,
	makeapp(Args1,A),
	checkhorpoge(A,T1,Right,T2,XV,P),!.
checkhorpo1Lam(lam(X,U),(TI1->_),Right,T2,XV,p(3-d,P)):-  %Case 3d interm Eta
	nonvar(T2),
	T2=(TI2->TO2),!,
	eqType(TI1,TI2),!,
	replace(X,NV,U,V),
	addapp(Right,T2,NV:TI2,A),
	checkhorpo(V,A:TO2,XV,P).


addapp(A1,T,A2,@(A1:T,A2)):-
	var(A1),!.
addapp(A1,_,A2,A):-
	A1=..[@|Args1],!,
	append(Args1,[A2],Args),
	A=..[@|Args].
addapp(A1,T,A2,@(A1:T,A2)):-!.


remXapp(A:T,A:T,_):-var(A),!.
remXapp(@(AT1,A2:_):_,NAT,X):-
    memberVar(A2,X),
    remXapp(AT1,NAT,X).
remXapp(A:T,A:T,_):-!.
    
memberVar(X,[Y|_]):-X==Y.
memberVar(X,[_|L]):-
    memberVar(X,L).

%%%%%%%%%%%%%%%%%%%%%%% Accessible subterms %%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% by now only one level

buildsubtacceq(A,A1):-A==A1,!.
buildsubtacceq(A:T1,SA:T2):-
    nonvar(T1),
    T1 \= (_ -> _), 
    nonvar(A),
    A=..[F|Args], F \== @, F \== lam,
    nth(_,Args,SA1:T21),
    SA1==SA,T2==T21,
    smallpos(T2,T2,T1).

smallpos(T2,_,T1):-
    var(T2),!,
    gtType(T1,T2).
smallpos((T21->T22),Ti2,T1):-!,
    smallpos(T21,Ti2,T1),
    smallpos(T22,Ti2,T1).
smallpos(T2,Ti2,T1):-
    ((gtType(T1,T2),!);(eqType(T1,T2),positive(T2,Ti2))).


positive(_,T1):-
    nonvar(T1),
    T1 \= (_ -> _),!.
%    occurtype(T2,T1),!.
positive(T2,T1):-
    nonvar(T1),
    T1 = (T11->T12),
    positive(T2,T12),
    negative(T2,T11).

negative(T2,T1):-
    nonvar(T1),
    T1 \= (_ -> _),
    \+(eqType(T2,T1)),!.
negative(T2,T1):-
    nonvar(T1),
    T1 = (T11->T12),
    negative(T2,T12),
    positive(T2,T11).

occurtype(T2,T1):- 
    eqType(T2,T1),!.
occurtype(T2,T1):- 
    nonvar(T1),
    T1=..[_|TArgs],
    occurtypelist(T2,TArgs).

occurtypelist(T2,[T1|_]):-
    occurtype(T2,T1),!.
occurtypelist(T2,[_|RT]):-
    occurtypelist(T2,RT).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



eqlamTerm(X:T1,Y:T2):-
	(var(X);var(Y)),!,X==Y, eqType(T1,T2),!.
eqlamTerm(lam(X,U):T1,lam(Y,V):T2):-!,
	eqType(T1,T2),
	replace(Y,X,V,W),
	eqlamTerm(U,W).
eqlamTerm(A:T1,B:T2):-
	A=..[F|Args1],
	B=..[F|Args2],!,
	eqType(T1,T2),
	eqlamTermList(Args1,Args2).

eqlamTermList([],[]).
eqlamTermList([A1|Args1],[A2|Args2]):-
	eqlamTerm(A1,A2),
	eqlamTermList(Args1,Args2).

gtPrecF(C,C1):-
    g_read(funcprec,LP),
    gtPrec(C,C1,LP).

eqPrecF(C,C1):-
    g_read(funcprec,LP),
    eqPrec(C,C1,LP).

gtPrecT(C,C1):-
    g_read(typeprec,LP),
    gtPrec(C,C1,LP).

eqPrecT(C,C1):-
    g_read(typeprec,LP),
    eqPrec(C,C1,LP).

gtPrec(C,C1,LP):-
    nonvar(C),nonvar(C1),
    member([C,_,LG],LP),!,
    member(C1,LG).

eqPrec(C,C1,_):-
    (var(C);var(C1)),!,
    C==C1,!.
eqPrec(C,C1,_):-
    C==C1,!.
eqPrec(C,C1,LP):-
    member([C,LE,_],LP),!,
    member(C1,LE).

flatApp(A,Args1):-
    A=..[@,A1|Args],!,
    flatApp(A1,LA1),
    append(LA1,Args,Args1).
flatApp(A,[A]).

geType(X,Y):-
    X==Y.
geType(X,_):-
    var(X),!,fail.
geType(X,Y):-
    X=..[->,I1,O1],!,
    (geType(O1,Y);
     (nonvar(Y),Y=..[->,I2,O2],eqType(I1,I2),geType(O1,O2))).
geType(X,Y):-
    X=..[C|Args],
    g_read(typecons,LTC),
    membereqsol([C,_],LTC),!,
    geTypeCons(C,Args,Y).


geTypeCons(C,Args,Y):-
    nonvar(Y),
    Y=..[C1|Args1],
    g_read(typecons,LTC),
    membereqsol([C1,_],LTC),!,
    (gtPrecT(C,C1) -> gtTypeConstlist(C,Args,Args1) ;
     (eqPrecT(C,C1) -> geTypeStat(C,Args,Args1) ;
      (member(T,Args),geType(T,Y)))).
geTypeCons(_,Args,Y):-
	member(T,Args),
	geType(T,Y).

gtTypeConstlist(_,_,[]).
gtTypeConstlist(C,Args,[A1|Args1]):-
	geTypeCons(C,Args,A1),
	gtTypeConstlist(C,Args,Args1).

gtType(X,Y):-
    X==Y,!,fail.
gtType(X,_):-
    var(X),!,fail.
gtType(X,Y):-
    X=..[->,I1,O1],!,
    (gtType(O1,Y);
     (nonvar(Y),Y=..[->,I2,O2],eqType(I1,I2),gtType(O1,O2))).
gtType(X,Y):-
    X=..[C|Args],
    g_read(typecons,LTC),
    membereqsol([C,_],LTC),!,
    gtTypeCons(C,Args,Y).


gtTypeCons(C,Args,Y):-
    nonvar(Y),
    Y=..[C1|Args1],
    g_read(typecons,LTC),
    membereqsol([C1,_],LTC),!,
    (gtPrecT(C,C1) -> gtTypeConstlist(C,Args,Args1) ;
     (eqPrecT(C,C1) -> gtTypeStat(C,Args,Args1) ;
      (member(T,Args),geType(T,Y)))).
gtTypeCons(_,Args,Y):-
	member(T,Args),
	geType(T,Y).

geTypeStat(C,Args,Args1):-
    eqTypeStat(C,Args,Args1),!.
geTypeStat(C,Args,Args1):-
    gtTypeStat(C,Args,Args1),!.

eqType(I1,I2):-(var(I1);var(I2)),!,I1=I2.
eqType(I1,I2):-
	I1=..[->|Args1],
	I2=..[->|Args2],
	eqTypeList(Args1,Args2).
eqType(I1,I2):-
	I1=..[C1|Args1],
	I2=..[C2|Args2],
	eqPrecT(C1,C2),
	eqTypeStat(C1,Args1,Args2).

eqTypeStat(C,Args1,Args2):-
    g_read(typelex,[L1,L2]),
    (member(C,L1);member(C,L2)),
    eqTypeList(Args1,Args2).
eqTypeStat(_,Args1,Args2):-
    eqTypeMul(Args1,Args2).

eqTypeList([],[]). 
eqTypeList([T1|Args1],[T2|Args2]):-
    eqType(T1,T2),
    eqTypeList(Args1,Args2). 


eqTypeMul([],[]). 
eqTypeMul([T1|Args1],Args2):-
    delete1(T2,Args2,Args3),
    eqType(T1,T2),!,
    eqTypeMul(Args1,Args3). 


gtTypeStat(C,Args1,Args2):-
    g_read(typelex,[L1,L2]),
    (member(C,L1);member(C,L2)),
    gtTypeLex(Args1,Args2,Args3),
    gtTypeConstlist(C,Args1,Args3).
gtTypeStat(_,Args1,Args2):-
    gtTypeMul(Args1,Args2).

gtTypeLex([T1|_],[T2|Args2],Args2):-
    gtType(T1,T2). 
gtTypeLex([T1|Args1],[T2|Args2],Args3):-
    eqType(T1,T2),
    gtTypeLex(Args1,Args2,Args3). 


gtTypeMul([_|_],[]). 
gtTypeMul(Args1,[T2|Args2]):-
    delete1(T1,Args1,Args3),
    eqType(T1,T2),!,
    gtTypeMul(Args3,Args2). 
gtTypeMul(Args1,[T2|Args2]):-
    member(T1,Args1),
    gtType(T1,T2),!,
    geTypeMul(Args1,Args2). 

geTypeMul([],[]). 
geTypeMul(Args1,[T2|Args2]):-
    delete1(T1,Args1,Args3),
    eqType(T1,T2),!,
    geTypeMul(Args3,Args2). 
geTypeMul(Args1,[T2|Args2]):-
    member(T1,Args1),
    gtType(T1,T2),!,
    geTypeMul(Args1,Args2). 



write_lamterm(T):-flatlamapp(T,FT),write(FT),!.

flatlamapp(T,T):-var(T),!.
flatlamapp('$VAR'(N),'$VAR'(N)):-!.
flatlamapp(T,FT):-
	T=..[@|_],!,
	flatApp(T,Args),
	flatlamapplist(Args,Args1),
	FT=..[@|Args1].
flatlamapp(T,FT):-
	T=..[lam|_],!,
	flatLam(T,Vars,Body),
	flatlamapp(Body,Body1),
	append(Vars,[Body1],Args),
	FT=..[lam|Args].
flatlamapp(T,FT):-
	T=..[F|Args],
	flatlamapplist(Args,Args1),
	FT=..[F|Args1].

flatlamapplist([],[]).
flatlamapplist([A|Args],[A1|Args1]):-
	flatlamapp(A,A1),
	flatlamapplist(Args,Args1).


flatLam(lam(X,U):_,[X|Vars],Body):-!,
	flatLam(U,Vars,Body).
flatLam(lam(X,U),[X|Vars],Body):-!,
	flatLam(U,Vars,Body).
flatLam(Body,[],Body).


delete1(X,[X|L],L).
delete1(X,[Y|L1],[Y|L2]):-
    delete1(X,L1,L2).

nthdelete1(N,X,[E|L1],[[]|L2]):-
	E==[],
	nthdelete1(N1,X,L1,L2),
	N is N1 + 1.
nthdelete1(1,X,[X|L],[[]|L]).
nthdelete1(N,X,[Y|L1],[Y|L2]):-
    nthdelete1(N1,X,L1,L2),
    N is N1 + 1.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


writeblanks(0):-!.
writeblanks(N):-
    write(' '),
    N1 is N-1,
    writeblanks(N1),!.

writeNth(1):- write('1st'),!.
writeNth(2):- write('2nd'),!.
writeNth(3):- write('3rd'),!.
writeNth(N):- N<21,write(N),write('th'),!.
writeNth(N):- 1 is N mod 10, write(N),write('st'),!.
writeNth(N):- 2 is N mod 10, write(N),write('nd'),!.
writeNth(N):- 3 is N mod 10, write(N),write('rd'),!.
writeNth(N):- write(N),write('th'),!.

prettyprint(N,p(1-a,NA,equal)):-!, %Case 1 (a) Subterm
	writeblanks(N),
	write('By case 1 (a), the '), writeNth(NA), 
	write(' argument is equal to the rhs.'),
	nl,!.
prettyprint(N,p(1-a,NA,subtacc)):-!, %Case 1 (a) Accessible Subterm
	writeblanks(N),
	write('By case 1 (a), the rhs is obtained from an accessible subterm of the '), writeNth(NA), 
	write(' argument of the lhs.'),
	nl,!.
prettyprint(N,p(1-a,NA,P)):-	%Case 1 (a) Subterm
	writeblanks(N),
	write('By case 1 (a), the '),
	writeNth(NA), write(' argument is greater than the rhs:'),nl,
	N1 is N+3,
	prettyprint(N1,P),!.

prettyprint(N,p(1-b-mul,LP)):-	%Case 1 (b) Mul 
	writeblanks(N),
	write('By case 1 (b; multiset):'),nl,
	N1 is N+3,
	prettyprintMul(N1,LP),!.
prettyprint(N,p(1-b-lr,LP)):-	%Case 1 (b) Lex(lr)
	writeblanks(N),
	write('By case 1 (b; left-to-right):'),nl,
	N1 is N+3,
	prettyprintLexLR(N1,LP),!.
prettyprint(N,p(1-b-rl,LP)):-	%Case 1 (b) Lex(rl)
	writeblanks(N),
	write('By case 1 (b; right-to-left):'),nl,
	N1 is N+3,
	prettyprintLexRL(N1,LP),!.
prettyprint(N,p(1-c,[])):-	%Case 1 (c) Prec or Func-App
	writeblanks(N),
	write('By case 1 (c).'),nl,!.
prettyprint(N,p(1-c,LP)):-	%Case 1 (c) Prec or Func-App
	writeblanks(N),
	write('By case 1 (c):'),nl,
	N1 is N+3,
	prettyprintList(N1,1,LP),!.

prettyprint(N,p(2-a,NA,equal)):-!, %Case 2 (a) SubtApp 
	writeblanks(N),
	write('By case 2 (a), the '), writeNth(NA), 
	write(' argument is equal to the rhs.'),
	nl,!.
prettyprint(N,p(2-a,NA,subtacc)):-!, %Case 2 (a) SubtApp accessible
	writeblanks(N),
	write('By case 2 (a), the rhs is obtained from an accessible subterm of the '), writeNth(NA), 
	write(' argument of the lhs.'),
	nl,!.
prettyprint(N,p(2-a,NA,P)):-	%Case 2 (a) SubtApp greater
	writeblanks(N),
	write('By case 2 (a), the '),
	writeNth(NA), write(' argument is greater than the rhs:'),nl,
	N1 is N+3,
	prettyprint(N1,P),!.
prettyprint(N,p(2-b,pf(2),LP)):- %Case 2 (b) MulApp
	writeblanks(N),
	write('By case 2 (b):'),nl,
	N1 is N+3,
	prettyprintMul(N1,LP),!.
prettyprint(N,p(2-c,P)):- %Case 2 (c) beta
	writeblanks(N),
	write('By case 2 (c), the beta-reduction of the lhs is greater than the rhs:'),nl,
	N1 is N+3,
	prettyprint(N1,P),!.
prettyprint(N,p(2-d,p(2-b,pf(2),LP))):- %Case 2 (d) +  interm Beta
	writeblanks(N),
	write('We build a term that beta-reduces (by case 2c) to the rhs by abstracting'),nl,
	writeblanks(N),
        write('this rhs over its last argument and show that it is smaller than the lhs'),nl,
	writeblanks(N),
        write('by case 2 (b):'),nl,
	N1 is N+3,
	prettyprintMul(N1,LP),!.
prettyprint(N,p(3-a,equal)):-  %Case  3 (a) SubtLam
	writeblanks(N),
	write('By case 3 (a), the body is equal to the rhs.'),
	nl,!.
prettyprint(N,p(3-a,P)):-  %Case  3 (a) SubtLam
	writeblanks(N),
	write('By case 3 (a), the body of the lhs is greter than the rhs:'),nl,
	N1 is N+3,
	prettyprint(N1,P),!.
prettyprint(N,p(3-b,P)):-  %Case 3 (b) MonLam
	writeblanks(N),
	write('By case 3 (b), the body of the lhs is greter than the body of the rhs:'),nl,
	N1 is N+3,
	prettyprint(N1,P),!.
prettyprint(N,p(3-c,P)):-  %Case 3 (c) Eta
	writeblanks(N),
	write('By case 3 (c), the eta-reduction of the lhs is greater than the rhs:'),nl,
	N1 is N+3,
	prettyprint(N1,P),!.
prettyprint(N,p(3-d,P)):-  %Case 3-d interm Eta
	writeblanks(N),
	write('The term that eta-reduces (in one step) to the rhs (by case 3c) and'),nl,
	writeblanks(N),
        write('is smaller than the lhs by case 3 (b)'),nl,
	N1 is N+3,
	prettyprint(N1,P),!.
prettyprint(N,p(4-a)):-		%Case 4 (a) Right Bound var
	writeblanks(N),
	write('By case 4 (a).'),nl,!.
prettyprint(N,p(4-b,P)):-		%Case 4 (b) RightLam
	writeblanks(N),
	write('By case 4 (b):'),nl,
	N1 is N+3,
	prettyprint(N1,P),!.

prettyprintLexLR(NI,[p(NA,subtacc),args,LP]):-!,
	writeblanks(NI),
	write('- The '), writeNth(NA),
	write(' argument of the rhs is obtained from an accessible subterm of the one of the rhs.'),nl,
	writeblanks(NI),
	write('For the the arguments of the rhs we have:'),nl,
        NI1 is NI+3,
	prettyprintList(NI1,1,LP).
prettyprintLexLR(NI,[p(NA,P),args,LP]):-!,
	writeblanks(NI),
	write('- The '), writeNth(NA),
	write(' argument of the lhs is greater than the one of the rhs:'),nl,
	NI1 is NI+3,
	prettyprint(NI1,P),
	writeblanks(NI),
	write('For the the arguments of the rhs we have:'),nl,
	prettyprintList(NI1,1,LP).
prettyprintLexLR(NI,[p(NA,equal)|LP]):-
	writeblanks(NI),
	write('- The '), writeNth(NA),
	write(' argument of the lhs is equal to the one of the rhs:'),nl,
	prettyprintLexLR(NI,LP),!.

prettyprintLexRL(NI,[p(NA,subtacc),args,LP]):-!,
	writeblanks(NI),
	write('- The '), writeLast(NA),
	write(' argument of the rhs is obtained from an accessible subterm of the one of the rhs:'),nl,
	writeblanks(NI),
	write('For the the arguments of the rhs we have:'),nl,
        NI1 is NI+3,
	prettyprintList(NI1,1,LP).
prettyprintLexRL(NI,[p(NA,P),args,LP]):-!,
	writeblanks(NI),
	write('- The '), writeLast(NA),
	write(' argument of the lhs is greater than the one of the rhs:'),nl,
	NI1 is NI+3,
	prettyprint(NI1,P),
	writeblanks(NI),
	write('For the arguments of the rhs we have:'),nl,
        reverse(LP,LP1),
	prettyprintList(NI1,1,LP1).
prettyprintLexRL(NI,[p(NA,equal)|LP]):-
	writeblanks(NI),
	write('- The '), writeLast(NA),
	write(' argument of the lhs is equal to the one of the rhs:'),nl,
	prettyprintLexRL(NI,LP),!.

writeLast(0):-write('last'),!.
writeLast(N):-write('last-'),write(N),!.

prettyprintMul(_,[]).
%prettyprintMul(NI,[p(N1,N,subtacc)]):-!,
%    writeblanks(NI),
%    write('- The '), writeNth(N),
%    write(' argument of the rhs is obtained from an accessible subterm of the '),
%    writeNth(N1), write(' element of the lhs.'), nl,!.
prettyprintMul(NI,[p(N1,N,equal)]):-!,
    writeblanks(NI),
    write('- The '),writeNth(N1), 
    write(' element of the lhs is equal to the '),
    writeNth(N), write(' element of the rhs.'), nl,!.
prettyprintMul(NI,[p(N1,N,P)]):-
    P\=subtacc,
    writeblanks(NI),
    write('- The '),writeNth(N1), 
    write(' element of the lhs is greater than '),
    writeNth(N), write(' element of the rhs.'), nl,
    NI1 is NI+3,
    prettyprint(NI1,P),!.
prettyprintMul(NI,[p(N1,N,subtacc),args,LP]):-!,
    writeblanks(NI),
    write('- The '), writeNth(N),
    write(' argument of the rhs is obtained from an accessible subterm of the '),
    writeNth(N1), write(' argument of the lhs.'), nl,
    writeblanks(NI),
    write('For the arguments of the rhs we have:'),nl,
    NI1 is NI+3,
    prettyprintList(NI1,1,LP),!.
prettyprintMul(NI,[p(N1,N,equal),args,LP]):-!,
    writeblanks(NI),
    write('- The '),writeNth(N1), 
    write(' element of the lhs is equal to the '),
    writeNth(N), write(' element of the rhs.'), nl,
    writeblanks(NI),
    write('For the arguments of the rhs we have:'),nl,
    NI1 is NI+3,
    prettyprintList(NI1,1,LP),!.
prettyprintMul(NI,[p(N1,N,P),args,LP]):-
    writeblanks(NI),
    write('- The '),writeNth(N1), 
    write(' element of the lhs is greater than '),
    writeNth(N), write(' element of the rhs.'), nl,
    NI1 is NI+3,
    prettyprint(NI1,P),
    writeblanks(NI),
    write('For the arguments of the rhs we have:'),nl,
    prettyprintList(NI1,1,LP),!.
prettyprintMul(NI,[p(N1,N,subtacc)|LP]):-!,
    writeblanks(NI),
    write('- The '), writeNth(N),
    write(' argument of the rhs is obtained from an accessible subterm of the '),
    writeNth(N1), write(' argument of the lhs.'), nl,
    prettyprintMul(NI,LP).
prettyprintMul(NI,[p(N1,N,equal)|LP]):-!,
    writeblanks(NI),
    write('- The '),writeNth(N1), 
    write(' element of the lhs is equal to the '),
    writeNth(N), write(' element of the rhs.'), nl,
    prettyprintMul(NI,LP).
prettyprintMul(NI,[p(N1,N,P)|LP]):-
    writeblanks(NI),
    write('- The '),writeNth(N1), 
    write(' element of the lhs is greater than '),
    writeNth(N), write(' element of the rhs.'), nl,
    NI1 is NI+3,
    prettyprint(NI1,P),
    prettyprintMul(NI,LP).

prettyprintList(_,_,[]). 
prettyprintList(NI,N,[p(P)|LP]):-
    writeblanks(NI),
    write('- For the '),writeNth(N),
    write(' argument of the rhs:'),nl,
    NI1 is NI+3,
    prettyprint(NI1,P),
    N1 is N+1,
    prettyprintList(NI,N1,LP).
